@@ -2,12 +2,13 @@
 #include "exceptions.h"
 #include "git-version.h"
 #include "lite_operations.h"
+#include "lock_enabled.h"
 #include "myutils.h"
 #include "operations.h"
 #include "platform.h"
-#include "streams.h"
 
 #include <cryptopp/cpu.h>
+#include <cryptopp/hmac.h>
 #include <cryptopp/osrng.h>
 #include <cryptopp/secblock.h>
 #include <fuse.h>
@@ -24,8 +25,6 @@
 #include <unordered_map>
 #include <vector>
 
-#include <fuse.h>
-
 #ifdef _MSC_VER
 #include <Windows.h>
 #else
@@ -37,12 +36,13 @@ using namespace securefs;
 namespace
 {
 
-const char* CONFIG_FILE_NAME = ".securefs.json";
+const char* const CONFIG_FILE_NAME = ".securefs.json";
 const unsigned MIN_ITERATIONS = 20000;
 const unsigned MIN_DERIVE_SECONDS = 1;
 const size_t CONFIG_IV_LENGTH = 32, CONFIG_MAC_LENGTH = 16;
-const char* PBKDF_ALGO_PKCS5 = "pkcs5-pbkdf2-hmac-sha256";
-const char* PBKDF_ALGO_SCRYPT = "scrypt";
+const char* const PBKDF_ALGO_PKCS5 = "pkcs5-pbkdf2-hmac-sha256";
+const char* const PBKDF_ALGO_SCRYPT = "scrypt";
+const char* const EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED = " ";
 
 const char* get_version_header(unsigned version)
 {
@@ -60,6 +60,13 @@ const char* get_version_header(unsigned version)
     }
 }
 
+std::unique_ptr<Json::CharReader> create_json_reader()
+{
+    Json::CharReaderBuilder builder;
+    builder["rejectDupKeys"] = true;
+    return std::unique_ptr<Json::CharReader>(builder.newCharReader());
+}
+
 enum class NLinkFixPhase
 {
     CollectingNLink,
@@ -237,9 +244,34 @@ void fix(const std::string& basedir, operations::FileSystemContext* fs)
     puts("Fix complete");
 }
 
+void maybe_derive_with_keyfile(const securefs::key_type& password_dervied_key,
+                               StringRef maybe_key_file_path,
+                               securefs::key_type& out_key)
+{
+    if (maybe_key_file_path.empty())
+    {
+        out_key = password_dervied_key;
+        return;
+    }
+    auto file_stream = OSService::get_default().open_file_stream(maybe_key_file_path, O_RDONLY, 0);
+    byte buffer[4096];
+    CryptoPP::HMAC<CryptoPP::SHA256> hmac(password_dervied_key.data(), password_dervied_key.size());
+    while (true)
+    {
+        auto sz = file_stream->sequential_read(buffer, sizeof(buffer));
+        if (sz <= 0)
+        {
+            break;
+        }
+        hmac.Update(buffer, sz);
+    }
+    hmac.TruncatedFinal(out_key.data(), out_key.size());
+}
+
 Json::Value generate_config(unsigned int version,
                             const std::string& pbkdf_algorithm,
                             const CryptoPP::AlignedSecByteBlock& master_key,
+                            StringRef maybe_key_file_path,
                             const securefs::key_type& salt,
                             const void* password,
                             size_t pass_len,
@@ -249,7 +281,7 @@ Json::Value generate_config(unsigned int version,
 {
     Json::Value config;
     config["version"] = version;
-    securefs::key_type key_to_encrypt;
+    securefs::key_type password_derived_key;
     CryptoPP::AlignedSecByteBlock encrypted_master_key(nullptr, master_key.size());
 
     if (pbkdf_algorithm == PBKDF_ALGO_PKCS5)
@@ -260,12 +292,12 @@ Json::Value generate_config(unsigned int version,
                                                            salt.size(),
                                                            rounds ? rounds : MIN_ITERATIONS,
                                                            rounds ? 0 : MIN_DERIVE_SECONDS,
-                                                           key_to_encrypt.data(),
-                                                           key_to_encrypt.size());
+                                                           password_derived_key.data(),
+                                                           password_derived_key.size());
     }
     else if (pbkdf_algorithm == PBKDF_ALGO_SCRYPT)
     {
-        uint32_t N = rounds > 0 ? rounds : 65536, r = 8, p = 1;
+        uint32_t N = rounds > 0 ? rounds : (1u << 18u), r = 8, p = 1;
         config["iterations"] = N;
         config["scrypt_r"] = r;
         config["scrypt_p"] = p;
@@ -276,8 +308,8 @@ Json::Value generate_config(unsigned int version,
                                    N,
                                    r,
                                    p,
-                                   key_to_encrypt.data(),
-                                   key_to_encrypt.size());
+                                   password_derived_key.data(),
+                                   password_derived_key.size());
     }
     else
     {
@@ -291,8 +323,11 @@ Json::Value generate_config(unsigned int version,
     byte mac[CONFIG_MAC_LENGTH];
     generate_random(iv, array_length(iv));
 
+    securefs::key_type wrapping_key;
+    maybe_derive_with_keyfile(password_derived_key, maybe_key_file_path, wrapping_key);
+
     CryptoPP::GCM<CryptoPP::AES>::Encryption encryptor;
-    encryptor.SetKeyWithIV(key_to_encrypt.data(), key_to_encrypt.size(), iv, array_length(iv));
+    encryptor.SetKeyWithIV(wrapping_key.data(), wrapping_key.size(), iv, array_length(iv));
     encryptor.EncryptAndAuthenticate(encrypted_master_key.data(),
                                      mac,
                                      array_length(mac),
@@ -319,6 +354,7 @@ Json::Value generate_config(unsigned int version,
 }
 
 bool parse_config(const Json::Value& config,
+                  StringRef maybe_key_file_path,
                   const void* password,
                   size_t pass_len,
                   CryptoPP::AlignedSecByteBlock& master_key,
@@ -347,7 +383,7 @@ bool parse_config(const Json::Value& config,
 
     byte iv[CONFIG_IV_LENGTH];
     byte mac[CONFIG_MAC_LENGTH];
-    key_type salt, key_to_encrypt_master_key;
+    key_type salt, password_derived_key;
     CryptoPP::AlignedSecByteBlock encrypted_key;
 
     std::string salt_hex = config["salt"].asString();
@@ -375,8 +411,8 @@ bool parse_config(const Json::Value& config,
                           salt.size(),
                           iterations,
                           0,
-                          key_to_encrypt_master_key.data(),
-                          key_to_encrypt_master_key.size());
+                          password_derived_key.data(),
+                          password_derived_key.size());
     }
     else if (pbkdf_algorithm == PBKDF_ALGO_SCRYPT)
     {
@@ -389,17 +425,18 @@ bool parse_config(const Json::Value& config,
                          iterations,
                          r,
                          p,
-                         key_to_encrypt_master_key.data(),
-                         key_to_encrypt_master_key.size());
+                         password_derived_key.data(),
+                         password_derived_key.size());
     }
     else
     {
         throw_runtime_error("Unknown pbkdf algorithm " + pbkdf_algorithm);
     }
 
+    securefs::key_type wrapping_key;
+    maybe_derive_with_keyfile(password_derived_key, maybe_key_file_path, wrapping_key);
     CryptoPP::GCM<CryptoPP::AES>::Decryption decryptor;
-    decryptor.SetKeyWithIV(
-        key_to_encrypt_master_key.data(), key_to_encrypt_master_key.size(), iv, array_length(iv));
+    decryptor.SetKeyWithIV(wrapping_key.data(), wrapping_key.size(), iv, array_length(iv));
     return decryptor.DecryptAndVerify(master_key.data(),
                                       mac,
                                       array_length(mac),
@@ -420,7 +457,10 @@ std::shared_ptr<FileStream> CommandBase::open_config_stream(const std::string& p
     return OSService::get_default().open_file_stream(path, flags, 0644);
 }
 
-FSConfig CommandBase::read_config(FileStream* stream, const void* password, size_t pass_len)
+FSConfig CommandBase::read_config(FileStream* stream,
+                                  const void* password,
+                                  size_t pass_len,
+                                  StringRef maybe_key_file_path)
 {
     FSConfig result;
 
@@ -435,14 +475,19 @@ FSConfig CommandBase::read_config(FileStream* stream, const void* password, size
         str.insert(str.end(), buffer, buffer + sz);
     }
 
-    Json::Reader reader;
     Json::Value value;
-    if (!reader.parse(str.data(), str.data() + str.size(), value))
-        throw_runtime_error(strprintf("Failure to parse the config file: %s",
-                                      reader.getFormattedErrorMessages().c_str()));
+    std::string error_message;
+    if (!create_json_reader()->parse(str.data(), str.data() + str.size(), &value, &error_message))
+        throw_runtime_error(
+            strprintf("Failure to parse the config file: %s", error_message.c_str()));
 
-    if (!parse_config(
-            value, password, pass_len, result.master_key, result.block_size, result.iv_size))
+    if (!parse_config(value,
+                      maybe_key_file_path,
+                      password,
+                      pass_len,
+                      result.master_key,
+                      result.block_size,
+                      result.iv_size))
         throw_runtime_error("Invalid password");
     result.version = value["version"].asUInt();
     return result;
@@ -462,6 +507,7 @@ static void copy_key(const CryptoPP::AlignedSecByteBlock& in_key, optional<key_t
 }
 
 void CommandBase::write_config(FileStream* stream,
+                               StringRef maybe_key_file_path,
                                const std::string& pbdkf_algorithm,
                                const FSConfig& config,
                                const void* password,
@@ -473,6 +519,7 @@ void CommandBase::write_config(FileStream* stream,
     auto str = generate_config(config.version,
                                pbdkf_algorithm,
                                config.master_key,
+                               maybe_key_file_path,
                                salt,
                                password,
                                pass_len,
@@ -483,7 +530,8 @@ void CommandBase::write_config(FileStream* stream,
     stream->sequential_write(str.data(), str.size());
 }
 
-class CommonCommandBase : public CommandBase
+/// A base class for all commands that require a data dir to be present.
+class _DataDirCommandBase : public CommandBase
 {
 protected:
     TCLAP::UnlabeledValueArg<std::string> data_dir{
@@ -495,13 +543,6 @@ protected:
         false,
         "",
         "config_path"};
-    TCLAP::ValueArg<std::string> pass{
-        "",
-        "pass",
-        "Password (prefer manually typing or piping since those methods are more secure)",
-        false,
-        "",
-        "password"};
 
 protected:
     std::string get_real_config_path()
@@ -511,17 +552,73 @@ protected:
     }
 };
 
+class _SinglePasswordCommandBase : public _DataDirCommandBase
+{
+protected:
+    TCLAP::ValueArg<std::string> pass{
+        "",
+        "pass",
+        "Password (prefer manually typing or piping since those methods are more secure)",
+        false,
+        "",
+        "password"};
+    TCLAP::ValueArg<std::string> keyfile{
+        "",
+        "keyfile",
+        "An optional path to a key file to use in addition to or in place of password",
+        false,
+        "",
+        "path"};
+    TCLAP::SwitchArg askpass{
+        "",
+        "askpass",
+        "When set to true, ask for password even if a key file is used. "
+        "password+keyfile provides even stronger security than one of them alone.",
+        false};
+    CryptoPP::AlignedSecByteBlock password;
+
+    void get_password(bool require_confirmation)
+    {
+        if (pass.isSet() && !pass.getValue().empty())
+        {
+            password.Assign(reinterpret_cast<const byte*>(pass.getValue().data()),
+                            pass.getValue().size());
+            CryptoPP::SecureWipeBuffer(reinterpret_cast<byte*>(&pass.getValue()[0]),
+                                       pass.getValue().size());
+            return;
+        }
+        if (keyfile.isSet() && !keyfile.getValue().empty() && !askpass.getValue())
+        {
+            password.Assign(reinterpret_cast<const byte*>(EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED),
+                            strlen(EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED));
+            return;
+        }
+        if (require_confirmation)
+        {
+            return OSService::read_password_with_confirmation("Enter password:", &password);
+        }
+        return OSService::read_password_no_confirmation("Enter password:", &password);
+    }
+
+    void add_all_args_from_base(TCLAP::CmdLine& cmd_line)
+    {
+        cmd_line.add(&data_dir);
+        cmd_line.add(&config_path);
+        cmd_line.add(&pass);
+        cmd_line.add(&keyfile);
+        cmd_line.add(&askpass);
+    }
+};
+
 static const std::string message_for_setting_pbkdf
     = strprintf("The algorithm to stretch passwords. Use %s for maximum protection (default), or "
                 "%s for compatibility with old versions of securefs",
                 PBKDF_ALGO_SCRYPT,
                 PBKDF_ALGO_PKCS5);
 
-class CreateCommand : public CommonCommandBase
+class CreateCommand : public _SinglePasswordCommandBase
 {
 private:
-    CryptoPP::AlignedSecByteBlock password;
-
     TCLAP::ValueArg<unsigned> rounds{
         "r",
         "rounds",
@@ -530,7 +627,7 @@ private:
         0,
         "integer"};
     TCLAP::ValueArg<unsigned int> format{
-        "", "format", "The filesystem format version (1,2,3)", false, 4, "integer"};
+        "", "format", "The filesystem format version (1,2,3,4)", false, 4, "integer"};
     TCLAP::ValueArg<unsigned int> iv_size{
         "", "iv-size", "The IV size (ignored for fs format 1)", false, 12, "integer"};
     TCLAP::ValueArg<unsigned int> block_size{
@@ -546,24 +643,14 @@ public:
     void parse_cmdline(int argc, const char* const* argv) override
     {
         TCLAP::CmdLine cmdline(help_message());
+        add_all_args_from_base(cmdline);
         cmdline.add(&iv_size);
         cmdline.add(&rounds);
-        cmdline.add(&data_dir);
-        cmdline.add(&config_path);
         cmdline.add(&format);
-        cmdline.add(&pass);
         cmdline.add(&store_time);
         cmdline.add(&block_size);
         cmdline.parse(argc, argv);
-
-        if (pass.isSet())
-        {
-            password.resize(pass.getValue().size());
-            memcpy(password.data(), pass.getValue().data(), password.size());
-            return;
-        }
-
-        OSService::read_password_with_confirmation("Password: ", &password);
+        get_password(true);
     }
 
     int execute() override
@@ -598,6 +685,7 @@ public:
             OSService::get_default().remove_file(get_real_config_path());
         });
         write_config(config_stream.get(),
+                     keyfile.getValue(),
                      pbkdf.getValue(),
                      config,
                      password.data(),
@@ -633,7 +721,7 @@ public:
     const char* help_message() const noexcept override { return "Create a new filesystem"; }
 };
 
-class ChangePasswordCommand : public CommonCommandBase
+class ChangePasswordCommand : public _DataDirCommandBase
 {
 private:
     CryptoPP::AlignedSecByteBlock old_password, new_password;
@@ -646,17 +734,53 @@ private:
         "integer"};
     TCLAP::ValueArg<std::string> pbkdf{
         "", "pbkdf", message_for_setting_pbkdf, false, PBKDF_ALGO_SCRYPT, "string"};
+    TCLAP::ValueArg<std::string> old_key_file{
+        "", "oldkeyfile", "Path to original key file", false, "", "path"};
+    TCLAP::ValueArg<std::string> new_key_file{
+        "", "newkeyfile", "Path to new key file", false, "", "path"};
+    TCLAP::SwitchArg askoldpass{
+        "",
+        "askoldpass",
+        "When set to true, ask for password even if a key file is used. "
+        "password+keyfile provides even stronger security than one of them alone.",
+        false};
+    TCLAP::SwitchArg asknewpass{
+        "",
+        "asknewpass",
+        "When set to true, ask for password even if a key file is used. "
+        "password+keyfile provides even stronger security than one of them alone.",
+        false};
 
 public:
     void parse_cmdline(int argc, const char* const* argv) override
     {
         TCLAP::CmdLine cmdline(help_message());
-        cmdline.add(&rounds);
         cmdline.add(&data_dir);
+        cmdline.add(&rounds);
         cmdline.add(&config_path);
+        cmdline.add(&old_key_file);
+        cmdline.add(&new_key_file);
+        cmdline.add(&askoldpass);
+        cmdline.add(&asknewpass);
         cmdline.parse(argc, argv);
-        OSService::read_password_no_confirmation("Old password: ", &old_password);
-        OSService::read_password_with_confirmation("New password: ", &new_password);
+        if (old_key_file.getValue().empty() || askoldpass.getValue())
+        {
+            OSService::read_password_no_confirmation("Old password: ", &old_password);
+        }
+        else
+        {
+            old_password.Assign(reinterpret_cast<const byte*>(EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED),
+                                strlen(EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED));
+        }
+        if (new_key_file.getValue().empty() || asknewpass.getValue())
+        {
+            OSService::read_password_with_confirmation("New password: ", &new_password);
+        }
+        else
+        {
+            new_password.Assign(reinterpret_cast<const byte*>(EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED),
+                                strlen(EMPTY_PASSWORD_WHEN_KEY_FILE_IS_USED));
+        }
     }
 
     int execute() override
@@ -666,11 +790,13 @@ public:
         generate_random(buffer, array_length(buffer));
         auto tmp_path = original_path + hexify(buffer, array_length(buffer));
         auto stream = OSService::get_default().open_file_stream(original_path, O_RDONLY, 0644);
-        auto config = read_config(stream.get(), old_password.data(), old_password.size());
+        auto config = read_config(
+            stream.get(), old_password.data(), old_password.size(), old_key_file.getValue());
         stream = OSService::get_default().open_file_stream(
             tmp_path, O_WRONLY | O_CREAT | O_EXCL, 0644);
         DEFER(if (std::uncaught_exception()) { OSService::get_default().remove_file(tmp_path); });
         write_config(stream.get(),
+                     new_key_file.getValue(),
                      pbkdf.getValue(),
                      config,
                      new_password.data(),
@@ -687,14 +813,13 @@ public:
 
     const char* help_message() const noexcept override
     {
-        return "Change password of existing filesystem";
+        return "Change password/keyfile of existing filesystem";
     }
 };
 
-class MountCommand : public CommonCommandBase
+class MountCommand : public _SinglePasswordCommandBase
 {
 private:
-    CryptoPP::AlignedSecByteBlock password;
     TCLAP::SwitchArg single_threaded{"s", "single", "Single threaded mode"};
     TCLAP::SwitchArg background{
         "b", "background", "Run securefs in the background (currently no effect on Windows)"};
@@ -713,21 +838,37 @@ private:
         "options"};
     TCLAP::UnlabeledValueArg<std::string> mount_point{
         "mount_point", "Mount point", true, "", "mount_point"};
-    TCLAP::SwitchArg case_insensitive{"",
-                                      "insensitive",
-                                      "Converts the case of all filenames so "
-                                      "that it works case insensitively"};
-    TCLAP::SwitchArg enable_nfc{
-        "",
-        "nfc",
-        "Normalizes all filenames to normal form composed (NFC). By default, enabled on macOS and "
-        "disabled on other platforms for compatibility of native applications.",
+    TCLAP::ValueArg<std::string> fsname{
+        "", "fsname", "Filesystem name shown when mounted", false, "securefs", "fsname"};
+    TCLAP::ValueArg<std::string> fssubtype{
+        "", "fssubtype", "Filesystem subtype shown when mounted", false, "securefs", "fssubtype"};
+    TCLAP::SwitchArg noflock{"",
+                             "noflock",
+                             "Disables the usage of file locking. Needed on some network "
+                             "filesystems. May cause data loss, so use it at your own risk!",
+                             false};
+    TCLAP::ValueArg<std::string> normalization{"",
+                                               "normalization",
+                                               "Mode of filename normalization. Valid values: "
+                                               "none, casefold, nfc, casefold+nfc. Defaults to nfc "
+                                               "on macOS and none on other platforms",
+                                               false,
 #ifdef __APPLE__
-        true
+                                               "nfc",
 #else
-        false
+                                               "none",
 #endif
-    };
+                                               ""};
+
+private:
+    std::vector<const char*> to_c_style_args(const std::vector<std::string>& args)
+    {
+        std::vector<const char*> result(args.size());
+        std::transform(args.begin(), args.end(), result.begin(), [](const std::string& s) {
+            return s.c_str();
+        });
+        return result;
+    }
 
 public:
     void parse_cmdline(int argc, const char* const* argv) override
@@ -737,37 +878,33 @@ public:
 #ifdef __APPLE__
         cmdline.add(&noxattr);
 #endif
-
+        add_all_args_from_base(cmdline);
         cmdline.add(&background);
         // cmdline.add(&insecure);
         cmdline.add(&verbose);
         cmdline.add(&trace);
         cmdline.add(&log);
-        cmdline.add(&data_dir);
-        cmdline.add(&config_path);
         cmdline.add(&mount_point);
-        cmdline.add(&pass);
         cmdline.add(&fuse_options);
         cmdline.add(&single_threaded);
-        cmdline.add(&case_insensitive);
-        cmdline.add(&enable_nfc);
+        cmdline.add(&normalization);
+        cmdline.add(&fsname);
+        cmdline.add(&fssubtype);
+        cmdline.add(&noflock);
         cmdline.parse(argc, argv);
 
-        if (pass.isSet() && !pass.getValue().empty())
-        {
-            password.resize(pass.getValue().size());
-            memcpy(password.data(), pass.getValue().data(), password.size());
-            generate_random(reinterpret_cast<byte*>(&pass.getValue()[0]), pass.getValue().size());
-        }
-        else
-        {
-            OSService::read_password_no_confirmation("Password: ", &password);
-        }
+        get_password(false);
 
         if (global_logger && verbose.getValue())
             global_logger->set_level(kLogVerbose);
         if (global_logger && trace.getValue())
             global_logger->set_level(kLogTrace);
+
+        set_lock_enabled(!noflock.getValue());
+        if (noflock.getValue() && !single_threaded.getValue())
+        {
+            WARN_LOG("Using --noflock without --single is highly dangerous");
+        }
     }
 
     void recreate_logger()
@@ -804,8 +941,8 @@ public:
         if (mount_point.getValue().size() != 2 || mount_point.getValue()[1] != ':'
             || !is_letter(mount_point.getValue().front()))
         {
-            ERROR_LOG("The mount point must be a drive path, such as Z:");
-            return 33;
+            WARN_LOG("The mount point on Windows should be a drive path, such as Z:, or some "
+                     "programs will get confused due to case sensitivity");
         }
 #else
         try
@@ -835,7 +972,8 @@ public:
             }
             throw;
         }
-        auto config = read_config(config_stream.get(), password.data(), password.size());
+        auto config = read_config(
+            config_stream.get(), password.data(), password.size(), keyfile.getValue());
         config_stream.reset();
         CryptoPP::SecureWipeBuffer(password.data(), password.size());
 
@@ -866,7 +1004,7 @@ public:
                      e.what());
         }
 
-        std::vector<const char*> fuse_args;
+        std::vector<std::string> fuse_args;
         fuse_args.push_back("securefs");
         if (config.version < 4 || single_threaded.getValue())
         {
@@ -893,15 +1031,15 @@ public:
         fuse_args.push_back("big_writes");
 #endif
         fuse_args.push_back("-o");
-        fuse_args.push_back("fsname=securefs");
+        fuse_args.push_back("fsname=" + fsname.getValue());
         fuse_args.push_back("-o");
-        fuse_args.push_back("subtype=securefs");
+        fuse_args.push_back("subtype=" + fssubtype.getValue());
         if (fuse_options.isSet())
         {
             for (const std::string& opt : fuse_options.getValue())
             {
                 fuse_args.push_back("-o");
-                fuse_args.push_back(opt.c_str());
+                fuse_args.push_back(opt);
             }
         }
 
@@ -927,12 +1065,31 @@ public:
         fsopt.flags = config.version < 3 ? 0 : kOptionStoreTime;
         if (insecure.getValue())
             fsopt.flags.value() |= kOptionNoAuthentication;
-        if (case_insensitive.getValue())
+        bool case_insensitive = false;
+        bool enable_nfc = false;
+        if (normalization.getValue() == "nfc")
+        {
+            enable_nfc = true;
+        }
+        else if (normalization.getValue() == "casefold")
+        {
+            case_insensitive = true;
+        }
+        else if (normalization.getValue() == "casefold+nfc")
+        {
+            case_insensitive = true;
+            enable_nfc = true;
+        }
+        else if (normalization.getValue() != "none")
+        {
+            throw_runtime_error("Invalid flag of --normalization: " + normalization.getValue());
+        }
+        if (case_insensitive)
         {
             INFO_LOG("Mounting as a case insensitive filesystem");
             fsopt.flags.value() |= kOptionCaseFoldFileName;
         }
-        if (enable_nfc.getValue())
+        if (enable_nfc)
         {
             INFO_LOG("Mounting as a Unicode normalized filesystem");
             fsopt.flags.value() |= kOptionNFCFileName;
@@ -993,7 +1150,7 @@ public:
         }
         recreate_logger();
         return fuse_main(static_cast<int>(fuse_args.size()),
-                         const_cast<char**>(fuse_args.data()),
+                         const_cast<char**>(to_c_style_args(fuse_args).data()),
                          &operations,
                          &fsopt);
     }
@@ -1005,26 +1162,23 @@ public:
     const char* help_message() const noexcept override { return "Mount an existing filesystem"; }
 };
 
-class FixCommand : public CommonCommandBase
+class FixCommand : public _SinglePasswordCommandBase
 {
-private:
-    CryptoPP::AlignedSecByteBlock password;
-
 public:
     void parse_cmdline(int argc, const char* const* argv) override
     {
         TCLAP::CmdLine cmdline(help_message());
-        cmdline.add(&data_dir);
-        cmdline.add(&config_path);
+        add_all_args_from_base(cmdline);
         cmdline.parse(argc, argv);
 
-        OSService::read_password_no_confirmation("Password: ", &password);
+        get_password(false);
     }
 
     int execute() override
     {
         auto config_stream = open_config_stream(get_real_config_path(), O_RDONLY);
-        auto config = read_config(config_stream.get(), password.data(), password.size());
+        auto config = read_config(
+            config_stream.get(), password.data(), password.size(), keyfile.getValue());
         config_stream.reset();
 
         if (config.version >= 4)
@@ -1168,11 +1322,17 @@ public:
 
         Json::Value config_json;
 
+        // Open a new scope to limit the lifetime of `buffer`.
         {
-            Json::Reader reader;
-            std::string buffer(fs->size(), 0);
-            fs->read(&buffer[0], 0, buffer.size());
-            reader.parse(buffer, config_json);
+            std::vector<char> buffer(fs->size(), 0);
+            fs->read(buffer.data(), 0, buffer.size());
+            std::string error_message;
+            if (!create_json_reader()->parse(
+                    buffer.data(), buffer.data() + buffer.size(), &config_json, &error_message))
+            {
+                throw_runtime_error(
+                    strprintf("Failure to parse the config file: %s", error_message.c_str()));
+            }
         }
 
         unsigned format_version = config_json["version"].asUInt();
