@@ -1,5 +1,5 @@
 #include "operations.h"
-#include "case_fold.h"
+#include "apple_xattr_workaround.h"
 #include "constants.h"
 #include "crypto.h"
 #include "platform.h"
@@ -61,8 +61,10 @@ namespace internal
 
     FileGuard open_base_dir(FileSystemContext* fs, const char* path, std::string& last_component)
     {
-        std::vector<std::string> components
-            = split((fs->flags & kOptionCaseFoldFileName) ? case_fold(path) : path, '/');
+        std::vector<std::string> components = split(
+            transform(path, fs->flags & kOptionCaseFoldFileName, fs->flags & kOptionNFCFileName)
+                .get(),
+            '/');
 
         FileGuard result(&fs->table, fs->table.open_as(fs->root_id, FileBase::DIRECTORY));
         if (components.empty())
@@ -209,10 +211,7 @@ namespace operations
 
     void* init(struct fuse_conn_info* fsinfo)
     {
-#ifdef FUSE_CAP_BIG_WRITES
-        fsinfo->want |= FUSE_CAP_BIG_WRITES;
-        fsinfo->max_write = static_cast<unsigned>(-1);
-#endif
+        (void)fsinfo;
         auto args = static_cast<MountOptions*>(fuse_get_context()->private_data);
         auto fs = new FileSystemContext(*args);
         TRACE_LOG("%s", __FUNCTION__);
@@ -674,13 +673,13 @@ namespace operations
         try
         {
             auto fg = internal::open_all(fs, path);
-            return static_cast<int>(fg->listxattr(list, size));
+            int rc = static_cast<int>(fg->listxattr(list, size));
+            transform_listxattr_result(list, size);
+            return rc;
         }
         COMMON_CATCH_BLOCK
     }
 
-    static const char* APPLE_FINDER_INFO = "com.apple.FinderInfo";
-
 #define XATTR_COMMON_PROLOGUE                                                                      \
     auto ctx = fuse_get_context();                                                                 \
     auto fs = internal::get_fs(ctx);                                                               \
@@ -708,6 +707,9 @@ namespace operations
 
         if (position != 0)
             return -EINVAL;
+        int rc = precheck_getxattr(&name);
+        if (rc <= 0)
+            return rc;
 
         try
         {
@@ -728,10 +730,9 @@ namespace operations
 
         if (position != 0)
             return -EINVAL;
-        if (strcmp(name, "com.apple.quarantine") == 0)
-            return 0;    // workaround for the "XXX is damaged" bug on OS X
-        if (strcmp(name, APPLE_FINDER_INFO) == 0)
-            return -EACCES;
+        int rc = precheck_setxattr(&name, &flags);
+        if (rc <= 0)
+            return rc;
 
         flags &= XATTR_CREATE | XATTR_REPLACE;
         try
@@ -746,6 +747,9 @@ namespace operations
     int removexattr(const char* path, const char* name)
     {
         XATTR_COMMON_PROLOGUE
+        int rc = precheck_removexattr(&name);
+        if (rc <= 0)
+            return rc;
 
         try
         {
@@ -777,12 +781,14 @@ namespace operations
         opt->release = &securefs::operations::release;
         opt->ftruncate = &securefs::operations::ftruncate;
         opt->flush = &securefs::operations::flush;
+#ifndef _WIN32
         opt->chmod = &securefs::operations::chmod;
         opt->chown = &securefs::operations::chown;
         opt->symlink = &securefs::operations::symlink;
-        opt->readlink = &securefs::operations::readlink;
-        opt->rename = &securefs::operations::rename;
         opt->link = &securefs::operations::link;
+        opt->readlink = &securefs::operations::readlink;
+#endif
+        opt->rename = &securefs::operations::rename;
         opt->fsync = &securefs::operations::fsync;
         opt->fsyncdir = &securefs::operations::fsyncdir;
         opt->utimens = &securefs::operations::utimens;
