@@ -1,6 +1,7 @@
 #ifndef WIN32
 #define _DARWIN_BETTER_REALPATH 1
 #include "exceptions.h"
+#include "lock_enabled.h"
 #include "logger.h"
 #include "platform.h"
 #include "streams.h"
@@ -57,6 +58,10 @@ public:
 
     void lock(bool exclusive) override
     {
+        if (!securefs::is_lock_enabled())
+        {
+            return;
+        }
         int rc = ::flock(m_fd, exclusive ? LOCK_EX : LOCK_SH);
         if (rc < 0)
         {
@@ -66,6 +71,10 @@ public:
 
     void unlock() noexcept override
     {
+        if (!securefs::is_lock_enabled())
+        {
+            return;
+        }
         int rc = ::flock(m_fd, LOCK_UN);
         (void)rc;
     }
@@ -77,7 +86,7 @@ public:
             THROW_POSIX_EXCEPTION(errno, "fsync");
     }
 
-    void fstat(struct stat* out) override
+    void fstat(struct stat* out) const override
     {
         if (!out)
             throwVFSException(EFAULT);
@@ -131,10 +140,9 @@ public:
 
     length_type size() const override
     {
-        off_t rc = ::lseek(m_fd, 0, SEEK_END);
-        if (rc < 0)
-            THROW_POSIX_EXCEPTION(errno, "lseek");
-        return static_cast<length_type>(rc);
+        struct stat st;
+        this->fstat(&st);
+        return static_cast<length_type>(st.st_size);
     }
 
     bool is_sparse() const noexcept override { return true; }
@@ -251,13 +259,21 @@ public:
     void rewind() override { ::rewinddir(m_dir); }
 };
 
-std::string OSService::norm_path(StringRef path) const
+bool OSService::is_absolute(StringRef path) { return path.size() > 0 && path[0] == '/'; }
+
+native_string_type OSService::concat_and_norm(StringRef base_dir, StringRef path)
 {
-    if (m_dir_fd < 0)
+    if (base_dir.empty() || is_absolute(path))
         return path.to_string();
-    if (path.size() > 0 && path[0] == '/')
-        return path.to_string();
-    return m_dir_name + path;
+    if (!is_absolute(base_dir))
+    {
+        throwInvalidArgumentException("base_dir must be absolute, but is " + base_dir);
+    }
+    if (base_dir.ends_with("/"))
+    {
+        return base_dir + path;
+    }
+    return base_dir + "/" + path;
 }
 
 OSService::OSService() { m_dir_fd = AT_FDCWD; }
@@ -310,6 +326,10 @@ void OSService::remove_directory(StringRef path) const
 
 void OSService::lock() const
 {
+    if (!securefs::is_lock_enabled())
+    {
+        return;
+    }
     int rc = ::flock(m_dir_fd, LOCK_NB | LOCK_EX);
     if (rc < 0)
         THROW_POSIX_EXCEPTION(errno,
