@@ -1,4 +1,5 @@
 #ifdef _WIN32
+#include "lock_enabled.h"
 #include "logger.h"
 #include "platform.h"
 
@@ -406,7 +407,8 @@ public:
     }
 };
 
-[[noreturn]] void throw_windows_exception(const wchar_t* funcname) {
+[[noreturn]] void throw_windows_exception(const wchar_t* funcname)
+{
     DWORD err = GetLastError();
     throw WindowsException(err, funcname);
 }
@@ -582,6 +584,10 @@ public:
 
     void lock(bool exclusive) override
     {
+        if (!securefs::is_lock_enabled())
+        {
+            return;
+        }
         OVERLAPPED o;
         memset(&o, 0, sizeof(o));
         CHECK_CALL(LockFileEx(m_handle,
@@ -594,6 +600,10 @@ public:
 
     void unlock() noexcept override
     {
+        if (!securefs::is_lock_enabled())
+        {
+            return;
+        }
         OVERLAPPED o;
         memset(&o, 0, sizeof(o));
         (void)(UnlockFileEx(
@@ -713,7 +723,7 @@ public:
         }
         CHECK_CALL(SetFileTime(m_handle, nullptr, &access_time, &mod_time));
     }
-    void fstat(struct fuse_stat* st) override { stat_file_handle(m_handle, st); }
+    void fstat(struct fuse_stat* st) const override { stat_file_handle(m_handle, st); }
     bool is_sparse() const noexcept override { return m_is_ntfs; }
 };
 
@@ -721,56 +731,64 @@ OSService::OSService() : m_root_handle(INVALID_HANDLE_VALUE) {}
 
 OSService::~OSService() { CloseHandle(m_root_handle); }
 
-std::wstring OSService::norm_path(StringRef path) const
+bool OSService::is_absolute(StringRef path)
 {
-    if (m_dir_name.empty() || path.empty()
-        || (path.size() > 0 && (path[0] == '/' || path[0] == '\\'))
-        || (path.size() > 2 && path[1] == ':'))
+    return path.empty() || path[0] == '/' || path[0] == '\\'
+        || (path.size() >= 2 && path[1] == ':');
+}
+
+native_string_type OSService::concat_and_norm(StringRef base_dir, StringRef path)
+{
+    if (base_dir.empty() || is_absolute(path))
+    {
         return widen_string(path);
+    }
+    if (!is_absolute(base_dir))
+    {
+        throwInvalidArgumentException("base_dir must be an absolute path, yet we received "
+                                      + base_dir);
+    }
+    std::string prepath = base_dir.to_string();
+    prepath.reserve(prepath.size() + 1 + path.size());
+    prepath.push_back('\\');
+    prepath.append(path.data(), path.size());
+    for (char& c : prepath)
+    {
+        if (c == '/')
+            c = '\\';
+    }
+    std::vector<std::string> components = split(prepath, '\\');
+    std::vector<const std::string*> norm_components;
+    norm_components.reserve(components.size());
+    for (const std::string& name : components)
+    {
+        if (name.empty() || name == ".")
+            continue;
+        if (name == "..")
+        {
+            if (norm_components.size() > 0)
+                norm_components.pop_back();
+            continue;
+        }
+        norm_components.push_back(&name);
+    }
+    std::string str;
+    str.reserve(prepath.size() + 8);
+    bool is_already_universal = prepath.size() >= 2 && prepath[0] == '\\' && prepath[1] == '\\';
+    if (is_already_universal)
+    {
+        str.push_back('\\');
+    }
     else
     {
-        std::string prepath = m_dir_name;
-        prepath.reserve(prepath.size() + 1 + path.size());
-        prepath.push_back('\\');
-        prepath.append(path.data(), path.size());
-        for (char& c : prepath)
-        {
-            if (c == '/')
-                c = '\\';
-        }
-        std::vector<std::string> components = split(prepath, '\\');
-        std::vector<const std::string*> norm_components;
-        norm_components.reserve(components.size());
-        for (const std::string& name : components)
-        {
-            if (name.empty() || name == ".")
-                continue;
-            if (name == "..")
-            {
-                if (norm_components.size() > 0)
-                    norm_components.pop_back();
-                continue;
-            }
-            norm_components.push_back(&name);
-        }
-        std::string str;
-        str.reserve(prepath.size() + 8);
-        bool is_already_universal = prepath.size() >= 2 && prepath[0] == '\\' && prepath[1] == '\\';
-        if (is_already_universal)
-        {
-            str.push_back('\\');
-        }
-        else
-        {
-            str.assign(("\\\\?"));
-        }
-        for (const std::string* name : norm_components)
-        {
-            str.push_back('\\');
-            str.append(*name);
-        }
-        return widen_string(str);
+        str.assign(("\\\\?"));
     }
+    for (const std::string* name : norm_components)
+    {
+        str.push_back('\\');
+        str.append(*name);
+    }
+    return widen_string(str);
 }
 
 OSService::OSService(StringRef path)
@@ -807,6 +825,10 @@ void OSService::remove_directory(StringRef path) const
 
 void OSService::lock() const
 {
+    if (!securefs::is_lock_enabled())
+    {
+        return;
+    }
     fprintf(stderr,
             "Warning: Windows does not support directory locking. "
             "Be careful not to mount the same data directory multiple times!\n");
@@ -1082,7 +1104,7 @@ std::string OSService::stringify_system_error(int errcode)
 void windows_init(void)
 {
     ::SetConsoleOutputCP(CP_UTF8);
-	// Use a large buffer to prevent Windows from chopping valid UTF-8 sequences.
+    // Use a large buffer to prevent Windows from chopping valid UTF-8 sequences.
     setvbuf(stdout, nullptr, _IOFBF, 65535);
     setvbuf(stderr, nullptr, _IOFBF, 65535);
     if (::FspLoad(nullptr) != STATUS_SUCCESS)
@@ -1206,4 +1228,4 @@ const char* PATH_SEPARATOR_STRING = "\\";
 const char PATH_SEPARATOR_CHAR = '\\';
 }    // namespace securefs
 
-#endif
+#endif
\ No newline at end of file
